#include <cassert>
#include <fstream>
#include <algorithm>

#ifdef DEBUG
#include <iostream>
#endif

#include <boost/system/error_code.hpp>
#include <boost/cstdint.hpp>
#include <boost/crc.hpp>

#include "message_history.h"

#define MESSAGES_FILENAME "messages.db"

namespace {

bool messageLess(const Message & a, const Message & b)
{
    return a.created() < b.created();
}

std::ostream & operator<<(std::ostream & stream, const Message & message)
{
    boost::crc_32_type crc32;
    crc32.process_bytes(message.text().c_str(), message.text().size());
    return stream << message.created() << " "
                  << message.received() << " "
                  << message.showTime() << " "
                  << message.type() << " "
                  << message.read() << " "
                  << message.text().size() << " "
                  << message.text() << " "
                  << crc32.checksum();
}

bool readMessage(std::istream & stream, Message & message)
{
    time_t created;
    time_t received;
    time_t showTime;
    unsigned type;
    bool read;
    size_t textSize;
    stream >> created >> received >> showTime >> type >> read >> textSize;
    if (stream.fail() ||
        stream.eof()) {
#ifdef DEBUG
        std::cerr << "Stream EOF or failure after reading message header" << std::endl;
#endif
        return false; // TODO: Boost error category
    }
    if (created == 0 ||
        received == 0 ||
        type == 0 ||
        textSize == 0) {
#ifdef DEBUG
        std::cerr << "Invalid message header (created: " << created << ", received: " << received << ", type: " << type << ", text size: " << textSize << ")" << std::endl;
#endif
        return false; // TODO: Boost error category
    }
    if (type > Message::TypesEnd ||
        textSize > 1024 * 1024 * 1024) {
#ifdef DEBUG
        std::cerr << "Invalid message header (created: " << created << ", received: " << received << ", type: " << type << ", text size: " << textSize << ")" << std::endl;
#endif
        return false; // TODO: Boost error category
    }
    char delim;
    stream.read(&delim, 1);
    std::vector<char> buffer(textSize);
    stream.read(&buffer.front(), buffer.size());
    if (stream.fail() ||
        stream.eof()) {
#ifdef DEBUG
        std::cerr << "Stream EOF or failure after reading message text" << std::endl;
#endif
        return false; // TODO: Boost error category
    }
    boost::uint32_t checksum;
    stream >> checksum;
    boost::crc_32_type crc32;
    crc32.process_bytes(&buffer.front(), buffer.size());
    if (crc32.checksum() != checksum) {
#ifdef DEBUG
        std::cerr << "Checksum verification failed (" << checksum << " != " << crc32.checksum() << ") for message '" << std::string(buffer.begin(), buffer.end()) << "'" << std::endl;
#endif
        return false; // TODO: Boost error category
    }
    stream.read(&delim, 1);
    message = Message(std::string(buffer.begin(), buffer.end()), created, received, showTime, type);
    if (read)
        message.markAsRead();
    return true;
}

class MessageSaver {
    public:
        MessageSaver(std::ostream & stream) : m_stream(stream) {}
        void operator()(const Message & message) { m_stream << message << "\n"; }

    private:
        std::ostream & m_stream;
};

}

MessageHistory::MessageHistory(const boost::filesystem::path & storagePath)
    : m_storagePath(storagePath),
      m_messages()
{
    restore(); // TODO: check errors
}

MessageHistory::~MessageHistory()
{
    save(); // TODO: check errors
}

size_t MessageHistory::add(const Message & message)
{
    const MessageList::iterator pos(
            std::lower_bound(
                m_messages.begin(),
                m_messages.end(),
                message,
                messageLess
            )
    );
    m_messages.insert(pos, message);
    return std::distance(m_messages.begin(), pos);
}

bool MessageHistory::erase(size_t pos)
{
    if (pos >= m_messages.size())
        return false; // TODO: Boost error category
    m_messages.erase(m_messages.begin() + pos);
    return true;
}

bool MessageHistory::eraseRange(size_t from, size_t count)
{
    if (from + count >= m_messages.size())
        return false; // TODO: Boost error category
    m_messages.erase(m_messages.begin() + from, m_messages.begin() + from + count);
    return true;
}

bool MessageHistory::save()
{
    const std::string tempFileName("."MESSAGES_FILENAME);
    const boost::filesystem::path tempFilePath(m_storagePath / tempFileName);
    {
        std::ofstream stream(tempFilePath.string().c_str());
        if (!stream)
            return false; // TODO: Boost error category
        std::for_each(m_messages.begin(), m_messages.end(), MessageSaver(stream));
    }
    const boost::filesystem::path filePath(m_storagePath / MESSAGES_FILENAME);
    boost::system::error_code ec;
    boost::filesystem::rename(tempFilePath, filePath, ec);
    if (ec)
        return false;
    return true;
}

bool MessageHistory::restore()
{
    MessageList messages;
    const boost::filesystem::path filePath(m_storagePath / MESSAGES_FILENAME);
    std::ifstream stream(filePath.string().c_str());
    if (!stream)
        return false; // TODO: Boost error category
    while (!stream.eof()) {
        Message message;
        if (!readMessage(stream, message))
            break;
        const MessageList::iterator pos(
                std::lower_bound(
                    messages.begin(),
                    messages.end(),
                    message,
                    messageLess
                )
        );
        messages.insert(pos, message);
    }
    m_messages.swap(messages);
#ifdef DEBUG
    std::cerr << "Messages history length: " << m_messages.size() << std::endl;
#endif
    return true;
}
