/*
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 1, or (at your option)
** any later version.

** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.

** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <iostream>

#include <boost/bind.hpp>

#include <QtCore>

#include <QApplication>
#include <QMessageBox>
#include <QSystemTrayIcon>
#include <QMenu>
#include <QKeySequence>

#include "ia_main_window.h"
#include "ia_settings_window.h"
#include "ia_tray_dialog.h"
#include "ia_quit_dialog.h"
#include "ia_about.h"

IAMainWindow::IAMainWindow(QSettings & s, const PathProvider & pp)
    : QWidget(NULL, 0),
      pathProvider(pp),
      logoIcon(pathProvider.getResourcePath("images/qia-logo.png").string().c_str()),
      onlineIcon(pathProvider.getResourcePath("images/ia_online.png").string().c_str()),
      offlineIcon(pathProvider.getResourcePath("images/ia_offline.png").string().c_str()),
      settings(s),
      iaFactory(settings.value("server_name", "192.168.1.1").toString().toStdString(),
                settings.value("server_port", "5555").toUInt(),
                settings.value("login", "").toString().toStdString(),
                settings.value("password", "").toString().toStdString(),
                settings.value("reconnect", "0").toBool()),
      disconnectedByUser(true),
      firstTime(true)
{
    setWindowIcon(logoIcon);
    setWindowTitle(tr("InetAccess"));
    // Layouts
    mainLayout = new QHBoxLayout(this);
    leftPanel = new QVBoxLayout();
    rightPanel = new QVBoxLayout();
    downPanel = new QGridLayout();
    // Left panel widgets
    connectBtn = new QPushButton(tr("Connect"), this);
    settingsBtn = new QPushButton(tr("Settings"), this);
    messagesButton = new QPushButton(tr("Messages"), this);
    // Right panel widgets
    stateGroup = new QGroupBox(tr("State"), this);
    stateTbl = new QTableView(this);
    ind = new IAIndicator(this);
    // Down panel widgets
    freeMbLbl = new QLabel(tr("Free Mb:"), this);
    cashLbl = new QLabel(tr("Cash:"), this);
    freeMbValueLbl = new QLabel(this);
    cashValueLbl = new QLabel(this);

    createActions();

    if (QSystemTrayIcon::isSystemTrayAvailable()) {
        createTray();
    } else {
        sysTrayIcon = NULL;
    }

    messagesWindow = new MessageHistoryWindow(NULL);
    messagesWindow->hide();

    // Model for table view
    statsModel = new IAStatsModel(this);
    statsModel->SetUnits(settings.value("units", "0").toInt());
    stateTbl->setModel(statsModel);

    messagesModel = new MessagesModel(this, pathProvider.getDataDir());
    messagesWindow->setModel(messagesModel);

    setMinimumSize(540, 240);
    resize(settings.value("width", "540").toInt(), settings.value("height", "240").toInt());
    setLayout(mainLayout);

    stateGroup->setLayout(rightPanel);

    mainLayout->addLayout(leftPanel, 0);
    mainLayout->addWidget(stateGroup, 1);

    leftPanel->addWidget(connectBtn);
    leftPanel->addWidget(settingsBtn);
    leftPanel->addWidget(messagesButton);
    leftPanel->addStretch();

    rightPanel->addWidget(stateTbl);
    rightPanel->addWidget(ind);
    rightPanel->addLayout(downPanel);

    downPanel->addWidget(freeMbLbl, 0, 0);
    downPanel->addWidget(cashLbl, 1, 0);
    downPanel->addWidget(freeMbValueLbl, 0, 1);
    downPanel->addWidget(cashValueLbl, 1, 1);

    stateTbl->setMinimumWidth(350);
    connectBtn->setCheckable(true);

    iaFactory.setStatusCb(boost::bind(&IAMainWindow::onStatusChanged, this, _1));
    iaFactory.setStatCb(boost::bind(&IAMainWindow::onStatChanged, this, _1));
    iaFactory.setMessageCb(boost::bind(&IAMainWindow::onInfo, this, _1, _2, _3, _4));
    iaFactory.setErrorCb(boost::bind(&IAMainWindow::onError, this, _1, _2));
    iaFactory.setDirNamesCb(boost::bind(&IAMainWindow::onDirNames, this, _1));

    // "Connect" action
    connect(connectBtn, SIGNAL(toggled(bool)),
            this, SLOT(toggleConnect(bool)));
    // "Settings" action
    connect(settingsBtn, SIGNAL(clicked()),
            this, SLOT(onSettings()));

    connect(messagesButton, SIGNAL(clicked()),
            this, SLOT(messagesButtonClicked()));

    connect(this, SIGNAL(stateChanged(bool)),
            this, SLOT(setState(bool)), Qt::QueuedConnection);
    // Action for various errors
    connect(this, SIGNAL(errorOccured(const QString &)),
            this, SLOT(onErrorOccured(const QString &)), Qt::QueuedConnection);

    connect(this, SIGNAL(incomingMessage()),
            messagesWindow, SLOT(showLast()), Qt::QueuedConnection);

    // autoconnect
    if (settings.value("autoconnect", "0").toBool()) {
        onConnect();
    }

    if (settings.value("minimize_on_close", "1").toBool()) {
        // Do not quit on window clode
        qApp->setQuitOnLastWindowClosed(false);
    } else {
        // Normal mode
        qApp->setQuitOnLastWindowClosed(true);
    }
}

IAMainWindow::~IAMainWindow()
{
    settings.setValue("width", size().width());
    settings.setValue("height", size().height());
}

// Slots
void IAMainWindow::onConnect()
{
    connectBtn->setText(tr("Disconnect"));
    if (!protoPtr)
        protoPtr.reset(iaFactory.newIA());
    protoPtr->connect();
    disconnectedByUser = false;
}

void IAMainWindow::onDisconnect()
{
    connectBtn->setText(tr("Connect"));
    if (!protoPtr)
        protoPtr.reset(iaFactory.newIA());
    protoPtr->disconnect();
    disconnectedByUser = true;
}

// Change authorization state
void IAMainWindow::toggleConnect(bool state)
{
    if (state) {
        onConnect();
    } else {
        onDisconnect();
    }
}

// Wrapper to show error messages
void IAMainWindow::onErrorOccured(const QString & errorMessage)
{
    //QMessageBox::critical(this, tr("Error"), errorMessage);
    QMessageBox::critical(0, tr("Error"), errorMessage);
}

// Show settings dialog
void IAMainWindow::onSettings()
{
    IASettingsWindow sw(this);

    // Setup form
    sw.setServerName(settings.value("server_name", "192.168.1.1").toString());
    sw.setServerPort(settings.value("server_port", "5555").toUInt());
    sw.setLogin(settings.value("login", "").toString());
    if (settings.value("link_password_to_hdd", "0").toBool()) {
        QString password(settings.value("password", "").toString());
        // TODO: Decrypt password
        sw.setPassword(password);
    } else {
        sw.setPassword(settings.value("password", "").toString());
    }
    sw.setSavePassword(settings.value("save_password", "1").toBool());
    sw.setLinkHDD(settings.value("link_password_to_hdd", "0").toBool());
    sw.setMinimize(settings.value("minimize_on_close", "1").toBool());
    sw.setMinimizeStart(settings.value("minimize_on_start","0").toBool());
    sw.setAutoconnect(settings.value("autoconnect", "0").toBool());
    sw.setReconnect(settings.value("reconnect", "0").toBool());
    sw.setUnits(settings.value("units", "0").toInt());
    sw.setDirNames(dirNames);

    if (sw.exec() == QDialog::Accepted) {
    // Update settings
        settings.setValue("server_name", sw.getServerName());
        settings.setValue("server_port", sw.getServerPort());
        settings.setValue("login", sw.getLogin());
        settings.setValue("save_password", sw.getSavePassword());
        settings.setValue("link_password_to_hdd", sw.getLinkHDD());
        if (sw.getSavePassword()) {
            QString password(sw.getPassword());
            if (sw.getLinkHDD()) {
                // TODO: Encrypt password
            }
            settings.setValue("password", password);
        }
        settings.setValue("autoconnect", sw.getAutoconnect());
        settings.setValue("reconnect", sw.getReconnect());
        settings.setValue("units", sw.getUnits());
        settings.setValue("minimize_on_start",sw.getMinimizeStart());

        iaFactory.setServerName(settings.value("server_name", "192.168.1.1").toString().toStdString());
        iaFactory.setPort(settings.value("server_port", "5555").toUInt());
        iaFactory.setLogin(settings.value("login", "").toString().toStdString());
        iaFactory.setPassword(settings.value("password", "").toString().toStdString());
        iaFactory.setReconnect(settings.value("reconnect", "0").toBool());

        protoPtr.reset(iaFactory.newIA());

        statsModel->SetUnits(settings.value("units", "0").toInt());

        // Change minimization options
        if (sw.getMinimize()) {
            qApp->setQuitOnLastWindowClosed(false);
            if (!settings.value("minimize_on_close").toBool()) {
                firstTime = true;
            }
        } else {
            onShow();
            qApp->setQuitOnLastWindowClosed(true);
        }
        settings.setValue("minimize_on_close", sw.getMinimize());
    }
}

void IAMainWindow::messagesButtonClicked()
{
    messagesWindow->show();
}

// Callback for status changes
void IAMainWindow::onStatusChanged(bool status)
{
#ifdef DEBUG
    std::cerr << "IAMainWindow::onStatusChanged - status = " << status << std::endl;
#endif

    emit stateChanged(status);
}

// Callback for info :)
void IAMainWindow::onStatChanged(const LOADSTAT & loadStat)
{
    statsModel->SetStats(loadStat);
    freeMbValueLbl->setText(statsModel->getFreeMb());
    cashValueLbl->setText(statsModel->getCash());
}

// Callback for messages :)
void IAMainWindow::onInfo(const std::string & message, int infoType, int showTime, int sendTime)
{
    assert(infoType > 0);
    assert(sendTime > 0);
    messagesModel->addMessage(Message(message,
                                      static_cast<time_t>(sendTime),
                                      time(NULL),
                                      static_cast<time_t>(showTime),
                                      static_cast<unsigned>(infoType)));
    emit incomingMessage();
}

// Callback for errors
void IAMainWindow::onError(const std::string & message, int errorCode)
{
    switch (errorCode) {
        case IA_SERVER_ERROR:
            emit stateChanged(false);
            emit errorOccured(QString::fromUtf8(message.c_str()));
            break;
        case IA_BIND_ERROR:
            emit stateChanged(false);
            emit errorOccured(tr("Network error"));
            break;
        case IA_FCNTL_ERROR:
            emit stateChanged(false);
            emit errorOccured(tr("Network error"));
            break;
        case IA_GETHOSTBYNAME_ERROR:
            emit stateChanged(false);
            emit errorOccured(tr("Network error"));
            break;
        case IA_SELECT_ERROR:
            emit stateChanged(false);
            emit errorOccured(tr("Network error"));
            break;
        case IA_RECV_ERROR:
            emit stateChanged(false);
            emit errorOccured(tr("Network error"));
            break;
        case IA_SEND_ERROR:
            emit stateChanged(false);
            emit errorOccured(tr("Network error"));
            break;
        case IA_TIMEOUT:
            emit stateChanged(false);
            emit errorOccured(tr("Server not responding"));
            break;
        default:
            emit stateChanged(false);
            emit errorOccured(tr("Unknown error: ") + QString::number(errorCode));
            break;
    }
}

// Callback for dir names
void IAMainWindow::onDirNames(const std::vector<std::string> & dn)
{
    std::vector<std::string>::const_iterator it;
    dirNames.clear();
    dirIndex.clear();
    int i = 0;
    for (it = dn.begin(); it != dn.end(); ++it) {
        if (*it != "") {
            QString dirName(QString::fromUtf8(it->c_str()));
            dirNames.append(dirName);
            dirIndex[dirName] = i;
        }
        ++i;
    }
    statsModel->SetDirNames(dirNames, dirIndex);
}

// Menu actions
void IAMainWindow::createActions()
{
    connectAction = new QAction(onlineIcon, tr("Connect"), this);
    connect(connectAction, SIGNAL(triggered()), this, SLOT(onConnect()));

    disconnectAction = new QAction(offlineIcon, tr("Disconnect"), this);
    disconnectAction->setEnabled(false);
    connect(disconnectAction, SIGNAL(triggered()), this, SLOT(onDisconnect()));

    settingsAction = new QAction(QIcon(pathProvider.getResourcePath("images/ia_settings.png").string().c_str()), tr("Settings"), this);
    connect(settingsAction, SIGNAL(triggered()), this, SLOT(onSettings()));

    hideAction = new QAction(QIcon(pathProvider.getResourcePath("images/ia_hide.png").string().c_str()), tr("Hide"), this);
    connect(hideAction, SIGNAL(triggered()), this, SLOT(onHide()));

    showAction = new QAction(QIcon(pathProvider.getResourcePath("images/ia_show.png").string().c_str()), tr("Show"), this);
    showAction->setEnabled(false);
    connect(showAction, SIGNAL(triggered()), this, SLOT(onShow()));

    aboutAction = new QAction(tr("About"), this);
    connect(aboutAction, SIGNAL(triggered()), SLOT(slot_about()));

    aboutQtAction = new QAction(tr("About Qt"), this);
    connect(aboutQtAction, SIGNAL(triggered()), SLOT(slot_aboutQt()));

    quitAction = new QAction(QIcon(pathProvider.getResourcePath("images/ia_quit.png").string().c_str()), tr("Quit"), this);
    quitAction->setShortcut(Qt::CTRL + Qt::Key_Q);
    quitAction->setShortcutContext(Qt::ApplicationShortcut);
    addAction(quitAction);
    connect(quitAction, SIGNAL(triggered()), this, SLOT(slot_quit()));
}

// Tray icon creation
void IAMainWindow::createTray()
{
    QMenu * trayMenu = new QMenu(this);
    trayMenu->addAction(connectAction);
    trayMenu->addAction(disconnectAction);
    trayMenu->addSeparator();
    trayMenu->addAction(settingsAction);
    trayMenu->addSeparator();
    trayMenu->addAction(hideAction);
    trayMenu->addAction(showAction);
    trayMenu->addSeparator();
    trayMenu->addAction(aboutAction);
    trayMenu->addAction(aboutQtAction);
    trayMenu->addSeparator();
    trayMenu->addAction(quitAction);

    sysTrayIcon = new IATrayIcon(onlineIcon, offlineIcon, this);
    sysTrayIcon->setContextMenu(trayMenu);
    sysTrayIcon->show();

    connect(sysTrayIcon, SIGNAL(activated(QSystemTrayIcon::ActivationReason)), this, SLOT(onTrayActivated(QSystemTrayIcon::ActivationReason)));
    sysTrayIcon->installEventFilter(this);
}

void IAMainWindow::closeEvent(QCloseEvent * event)
{
    if (settings.value("minimize_on_close", "1").toBool() && sysTrayIcon != NULL) {
        onHide();
        if (settings.value("show_minimize_dialog", "true").toBool()) {
            ia_tray_dialog dia(this);
            dia.exec();
            if (dia.getstate() == true) {
                settings.setValue("show_minimize_dialog", "false");
            } else {
                settings.setValue("show_minimize_dialog", "true");
            }
        }
        event->ignore();
    } else {
        event->accept();
    }
}

// Hide window
void IAMainWindow::onHide()
{
    hideAction->setEnabled(false);
    hide();
    showAction->setEnabled(true);
}

// Show window
void IAMainWindow::onShow()
{
    showAction->setEnabled(false);
    show();
    hideAction->setEnabled(true);
}

// Tray events
void IAMainWindow::onTrayActivated(QSystemTrayIcon::ActivationReason reason)
{
    switch (reason) {
        case QSystemTrayIcon::Trigger:
        case QSystemTrayIcon::DoubleClick:
            if (isHidden()) {
                onShow();
            } else {
                onHide();
            }
            break;
        case QSystemTrayIcon::MiddleClick:
            break;
        case QSystemTrayIcon::Context:
            break;
        case QSystemTrayIcon::Unknown:
            break;
    }

}

void IAMainWindow::setState(bool state)
{
    ind->setState(state);
    if (state) {
        connectAction->setEnabled(false);
        disconnectAction->setEnabled(true);
        if (!connectBtn->isChecked()) {
            // Newer to be here, just protection
            connectBtn->setText(tr("Disconnect"));
            connectBtn->setChecked(true);
            connectAction->setEnabled(false);
            disconnectAction->setEnabled(true);
        }
    } else {
        connectAction->setEnabled(true);
        disconnectAction->setEnabled(false);
        if (connectBtn->isChecked()) {
            // On timeout
            connectBtn->setText(tr("Connect"));
            connectBtn->setChecked(false);
            connectAction->setEnabled(true);
            disconnectAction->setEnabled(false);
        }
    }
    if (sysTrayIcon != NULL) {
        sysTrayIcon->setState(state);
    }
}

void IAMainWindow::slot_quit(){
    if(settings.value("show_quit_dialog", "true").toBool()){
        ia_quit_dialog dia(this);
        if (dia.exec()) {
            // quit
            if (dia.getstate() == true){
                settings.setValue("show_quit_dialog", "false");
            } else {
                settings.setValue("show_quit_dialog", "true");
            }
            qApp->quit();
        } else {
            //cancel quit
            if (dia.getstate() == true){
                settings.setValue("show_quit_dialog", "false");
            } else {
                settings.setValue("show_quit_dialog", "true");
            }
        }
    } else {
        qApp->quit();
    }
}

void IAMainWindow::slot_about(){
    ia_about dia(this);
    dia.exec();
}

void IAMainWindow::slot_aboutQt(){
    QMessageBox::aboutQt(this, tr("InetAccess - About Qt"));
}

bool IAMainWindow::eventFilter(QObject *, QEvent * ev){
    // print tray toolTip
    if (ev->type() == QEvent::ToolTip) {
        sysTrayIcon->setToolTip("HELLO!");
    }
    return true;
}
